@@ -1,6 +1,5 @@
 #############################################################################
 #
-#	$Id: protocol.py,v 2.94.2.40 2010/05/22 16:31:10 irmen Exp $
 #	Pyro Protocol Adapters
 #
 #	This is part of "Pyro" - Python Remote Objects
@@ -8,6 +7,7 @@
 #
 #############################################################################
 
+from __future__ import with_statement
 import socket, struct, os, time, sys, hmac, types, random, errno, select
 import imp, marshal, new, __builtin__
 try:
@@ -175,6 +175,26 @@ def set_sock_keepalive(sock):
 		except:
 			Pyro.config.PYRO_SOCK_KEEPALIVE=0    # it didn't work--disable keepalives.
 
+# set socket to not inherit in subprocess
+try:
+	import fcntl
+	def set_sock_no_inherit(sock):
+		# Mark the given socket fd as non-inheritable (posix)
+		fd = sock.fileno() 
+		flags = fcntl.fcntl(fd, fcntl.F_GETFD) 
+		fcntl.fcntl(fd, fcntl.F_SETFD, flags | fcntl.FD_CLOEXEC)
+except ImportError:
+	# no fcntl available, try the windows version
+	try: 
+		from ctypes import windll, WinError 
+		def set_sock_no_inherit(sock): 
+			# mark the given socket fd as non-inheritable (Windows).
+			if not windll.kernel32.SetHandleInformation(sock.fileno(), 1, 0): 
+				raise WinError() 	
+	except ImportError:
+		# nothing available, define a dummy function
+		def set_sock_no_inherit(sock):
+			pass
 
 #------ PYRO: adapter (default Pyro wire protocol)
 #------ This adapter is for protocol version 4 ONLY
@@ -237,8 +257,7 @@ class PYROAdapter(object):
 		if URI.protocol not in ('PYRO', 'PYROLOC'):
 			Log.error('PYROAdapter','incompatible protocol in URI:',URI.protocol)
 			raise ProtocolError('incompatible protocol in URI')
-		try:
-			self.lock.acquire()   # only 1 thread at a time can bind the URI
+		with self.lock:   # only 1 thread at a time can bind the URI
 			try:
 				self.URI=URI
 				sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
@@ -269,8 +288,6 @@ class PYROAdapter(object):
 			except socket.error:
 				Log.msg('PYROAdapter','connection failed to URI',str(URI))
 				raise ProtocolError('connection failed')
-		finally:
-			self.lock.release()
 
 	def resolvePYROLOC_URI(self, newProtocol):
 		# This method looks up the object URI referenced by
@@ -433,71 +450,87 @@ class PYROAdapter(object):
 		return answer
 
 	def remoteInvocation(self, method, flags, *args):
-		try:
-			self.lock.acquire() # only 1 thread at a time may use this connection to call a remote method
-			return self._remoteInvocation(method, flags, *args)
-		finally:
-			self.lock.release()
+		with self.lock:
+			# only 1 thread at a time may use this connection to call a remote method
+			try:
+				self.__pyrocallbusy=True
+				return self._remoteInvocation(method, flags, *args)
+				self.__pyrocallbusy=False
+			finally:
+				if self.__pyrocallbusy:
+					# the call has been aborted before completion, close the connection
+					# to avoid corrupt transfers on the next call
+					self.release()
 
 	def _remoteInvocation(self, method, flags, *args):
 		if 'conn' not in self.__dict__.keys():
-			Log.msg('PYROAdapter','no connection, trying to bind again')
-			if 'URI' in self.__dict__.keys():
-				self.bindToURI(self.URI)
-			else:
-				raise ProtocolError('trying to rebind, but was never bound before')
+			Log.error('PYROAdapter','no connection available in remoteinvocation')
+			raise ProtocolError('no connection available in remoteinvocation')
 		if method in self.onewayMethods:
 			flags |= Pyro.constants.RIF_Oneway
 		body=pickle.dumps((self.URI.objectID,method,flags,args),Pyro.config.PYRO_PICKLE_FORMAT)
-		sock_sendmsg(self.conn.sock, self.createMsg(body), self.timeout)
-		if flags & Pyro.constants.RIF_Oneway:
-			return None		# no answer required, return immediately
-		ver,answer,pflags = self.receiveMsg(self.conn,1)  # read the server's response, send no further replies
-		if answer is None:
-			raise ProtocolError('incorrect answer received')
-
-		# Try to get the answer from the server.
-		# If there are import problems, try to get those modules from
-		# the server too (if mobile code is enabled).
-		if not Pyro.config.PYRO_MOBILE_CODE:
-			answer = pickle.loads(answer)
+		try:
+			sock_sendmsg(self.conn.sock, self.createMsg(body), self.timeout)
+		except (socket.error, ProtocolError, KeyboardInterrupt):
+			# Communication error during write. To avoid corrupt transfers, we close the connection.
+			# Otherwise we might receive the previous reply as a result of a new methodcall! 
+			# Special case for keyboardinterrupt: people pressing ^C to abort the client
+			# may be catching the keyboardinterrupt in their code. We should probably be on the
+			# safe side and release the proxy connection in this case too, because they might
+			# be reusing the proxy object after catching the exception...
+			self.release()
+			raise
 		else:
-			importer=None
-			try:
-				imp.acquire_lock()
-				loaded = 0
-				# XXX maxtries here...
-				while not loaded:
-					# install a custom importer to intercept any extra needed modules
-					# when unpickling the answer just obtained from the server
-					imp.acquire_lock()
-					importer = agent_import(__builtin__.__import__)
-					__builtin__.__import__ = importer
-					imp.release_lock()
- 
-					try:
-						answer = pickle.loads(answer)
-						loaded = 1
-					except ImportError:
-						mname = importer.name
-						if importer is not None:
-							__builtin__.__import__ = importer.orig_import
-							importer = None
-							self._retrieveCode(mname, 0)
- 
-			finally:
-				if importer is not None:
-					__builtin__.__import__ = importer.orig_import
-				imp.release_lock()
-
-		if isinstance(answer,PyroExceptionCapsule):
-			if isinstance(answer.excObj,_InternalNoModuleError):
-				# server couldn't load the module, send it
-				return self.processMissingModuleError(answer.excObj, method, flags, args)
+			if flags & Pyro.constants.RIF_Oneway:
+				self.__pyrocallbusy=False
+				return None		# no answer required, return immediately
+			ver,answer,pflags = self.receiveMsg(self.conn,1)  # read the server's response, send no further replies
+			self.__pyrocallbusy=False
+			if answer is None:
+				raise ProtocolError('incorrect answer received')
+	
+			# Try to get the answer from the server.
+			# If there are import problems, try to get those modules from
+			# the server too (if mobile code is enabled).
+			if not Pyro.config.PYRO_MOBILE_CODE:
+				answer = pickle.loads(answer)
 			else:
-				# we have an encapsulated exception, raise it again.
-				answer.raiseEx()
-		return answer
+				importer=None
+				try:
+					imp.acquire_lock()
+					loaded = 0
+					# XXX maxtries here...
+					while not loaded:
+						# install a custom importer to intercept any extra needed modules
+						# when unpickling the answer just obtained from the server
+						imp.acquire_lock()
+						importer = agent_import(__builtin__.__import__)
+						__builtin__.__import__ = importer
+						imp.release_lock()
+	 
+						try:
+							answer = pickle.loads(answer)
+							loaded = 1
+						except ImportError:
+							mname = importer.name
+							if importer is not None:
+								__builtin__.__import__ = importer.orig_import
+								importer = None
+								self._retrieveCode(mname, 0)
+	 
+				finally:
+					if importer is not None:
+						__builtin__.__import__ = importer.orig_import
+					imp.release_lock()
+	
+			if isinstance(answer,PyroExceptionCapsule):
+				if isinstance(answer.excObj,_InternalNoModuleError):
+					# server couldn't load the module, send it
+					return self.processMissingModuleError(answer.excObj, method, flags, args)
+				else:
+					# we have an encapsulated exception, raise it again.
+					answer.raiseEx()
+			return answer
 
 	def processMissingModuleError(self, errorinfo, method, flags, args):
 		# server couldn't load module, supply it
@@ -546,45 +579,55 @@ class PYROAdapter(object):
 		
 	# (private) receives a socket message, returns: (protocolver, message, protocolflags)
 	def receiveMsg(self,conn,noReply=0):
-		msg=sock_recvmsg(conn.sock, self.headerSize, self.timeout)
-		(hid, ver, hsiz, bsiz, pflags, crc) = struct.unpack(self.headerFmt,msg)
-		# store in the connection what pickle method this is
-		if pflags&PFLG_XMLPICKLE_GNOSIS:
-			conn.pflags|=PFLG_XMLPICKLE_GNOSIS
-		if ver!=self.version:
-			msg='incompatible protocol version'
-			Log.error('PYROAdapter',msg)
-			if not noReply:
-				# try to report error to client, but most likely the connection will terminate:
-				self.returnException(conn, ProtocolError(msg))
-			raise ProtocolError(msg)
-		if hid!=self.headerID or hsiz!=self.headerSize:
-			msg='invalid header'
-			Log.error('PYROAdapter',msg)
-			Log.error('PYROAdapter','INVALID HEADER DETAILS: ',conn,( hid, ver, hsiz, bsiz,pflags))
-			if not noReply:
-				# try to report error to client, but most likely the connection will terminate:
-				self.returnException(conn, ProtocolError(msg), shutdown=1)
-			raise ProtocolError(msg)
-		body=sock_recvmsg(conn.sock, bsiz, self.timeout)
-		if pflags&PFLG_CHECKSUM:
-			if _has_compression:
-				if crc!=zlib.adler32(body):
-					msg='checksum error'
-					Log.error('PYROAdapter',msg)
-					if not noReply:
-						self.returnException(conn, ProtocolError(msg))
-					raise ProtocolError(msg)
-			else:
-				raise ProtocolError('cannot perform checksum')
-		if pflags&PFLG_COMPRESSED:
-			if _has_compression:
-				body=zlib.decompress(body)
-			else:
-				# We received a compressed message but cannot decompress.
-				# Is this really a server error? We now throw an exception on the server...
-				raise ProtocolError('compression not supported')
-		return ver,body,pflags
+		try:
+			msg=sock_recvmsg(conn.sock, self.headerSize, self.timeout)
+			(hid, ver, hsiz, bsiz, pflags, crc) = struct.unpack(self.headerFmt,msg)
+			# store in the connection what pickle method this is
+			if pflags&PFLG_XMLPICKLE_GNOSIS:
+				conn.pflags|=PFLG_XMLPICKLE_GNOSIS
+			if ver!=self.version:
+				msg='incompatible protocol version'
+				Log.error('PYROAdapter',msg)
+				if not noReply:
+					# try to report error to client, but most likely the connection will terminate:
+					self.returnException(conn, ProtocolError(msg))
+				raise ProtocolError(msg)
+			if hid!=self.headerID or hsiz!=self.headerSize:
+				msg='invalid header'
+				Log.error('PYROAdapter',msg)
+				Log.error('PYROAdapter','INVALID HEADER DETAILS: ',conn,( hid, ver, hsiz, bsiz,pflags))
+				if not noReply:
+					# try to report error to client, but most likely the connection will terminate:
+					self.returnException(conn, ProtocolError(msg), shutdown=1)
+				raise ProtocolError(msg)
+			body=sock_recvmsg(conn.sock, bsiz, self.timeout)
+			if pflags&PFLG_CHECKSUM:
+				if _has_compression:
+					if crc!=zlib.adler32(body):
+						msg='checksum error'
+						Log.error('PYROAdapter',msg)
+						if not noReply:
+							self.returnException(conn, ProtocolError(msg))
+						raise ProtocolError(msg)
+				else:
+					raise ProtocolError('cannot perform checksum')
+			if pflags&PFLG_COMPRESSED:
+				if _has_compression:
+					body=zlib.decompress(body)
+				else:
+					# We received a compressed message but cannot decompress.
+					# Is this really a server error? We now throw an exception on the server...
+					raise ProtocolError('compression not supported')
+			return ver,body,pflags
+		except (socket.error, ProtocolError, KeyboardInterrupt),x:
+			# Communication error during read. To avoid corrupt transfers, we close the connection.
+			# Otherwise we might receive the previous reply as a result of a new methodcall! 
+			# Special case for keyboardinterrupt: people pressing ^C to abort the client
+			# may be catching the keyboardinterrupt in their code. We should probably be on the
+			# safe side and release the proxy connection in this case too, because they might
+			# be reusing the proxy object after catching the exception...
+			self.release() 
+			raise
 
 	def _unpickleRequest(self, pflags, body):
 		if pflags&PFLG_XMLPICKLE_GNOSIS:
@@ -612,19 +655,24 @@ class PYROAdapter(object):
 		#  (object ID, method name, flags, (arg1,arg2,...))
 		importer=fromlist=None
 		try:
-			try:
+			if Pyro.config.PYRO_MOBILE_CODE:
 				# install a custom importer to intercept any extra needed modules
 				# when unpickling the request just obtained from the client
-				imp.acquire_lock()
-				importer=agent_import(__builtin__.__import__)
-				__builtin__.__import__=importer
+				try:
+					imp.acquire_lock()
+					importer=agent_import(__builtin__.__import__)
+					__builtin__.__import__=importer
+					req=self._unpickleRequest(pflags, body)
+				finally:
+					__builtin__.__import__=importer.orig_import
+					imp.release_lock()
+			else:
+				# no mobile code; just unpickle the stuff without a custom importer.
 				req=self._unpickleRequest(pflags, body)
-				if type(req)!=tuple or len(req)!=4 or type(req[3])!=tuple:
-					# sanity check failed
-					raise ProtocolError("invalid request data format")
-			finally:
-				__builtin__.__import__=importer.orig_import
-				imp.release_lock()
+
+			if type(req)!=tuple or len(req)!=4 or type(req[3])!=tuple:
+				# sanity check failed
+				raise ProtocolError("invalid request data format")
 
 		except ImportError,x:
 			if Pyro.config.PYRO_MOBILE_CODE:
@@ -808,8 +856,7 @@ class PYROSSLAdapter(PYROAdapter):
 		if URI.protocol not in ('PYROSSL','PYROLOCSSL'):
 			Log.error('PYROSSLAdapter','incompatible protocol in URI:',URI.protocol)
 			raise ProtocolError('incompatible protocol in URI')
-		try:
-			self.lock.acquire()   # only 1 thread at a time can bind the URI
+		with self.lock:   # only 1 thread at a time can bind the URI
 			try:
 				self.URI=URI
 				sock = SSL.Connection(self.ctx,socket.socket(socket.AF_INET, socket.SOCK_STREAM))
@@ -835,8 +882,6 @@ class PYROSSLAdapter(PYROAdapter):
 			except socket.error:
 				Log.msg('PYROSSLAdapter','connection failed to URI',str(URI))
 				raise ProtocolError('connection failed')
-		finally:
-			self.lock.release()
 
 	def _sendConnect(self, sock, ident):
 		return PYROAdapter._sendConnect(self, sock, ident)
@@ -986,6 +1031,7 @@ class TCPServer(object):
 			# create server socket for new connections
 			self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
 			set_reuse_addr(self.sock)
+			set_sock_no_inherit(self.sock)
 			self.sock.bind((host,port))
 			self.sock.listen(Pyro.config.PYRO_TCP_LISTEN_BACKLOG)
 			if self._ssl_server:
